"""Tests for postprocessing modules."""

from __future__ import annotations

import ast

from stubgen_pyx.postprocessing import (
    collect_names,
    deduplicate_imports,
    normalize_names,
    sort_imports,
    trim_imports,
)


class TestCollectNames:
    """Test the collect_names module."""

    def test_collect_names_empty_module(self):
        """Test collecting names from an empty module."""
        tree = ast.parse("")
        names = collect_names.collect_names(tree)
        assert isinstance(names, set)

    def test_collect_names_simple_assignment(self):
        """Test collecting names from a simple assignment."""
        code = "x = 5"
        tree = ast.parse(code)
        names = collect_names.collect_names(tree)
        assert isinstance(names, set)

    def test_collect_names_function_def(self):
        """Test collecting names from a function definition."""
        code = """
def greet(name: str) -> str:
    return f"Hello, {name}"
"""
        tree = ast.parse(code)
        names = collect_names.collect_names(tree)
        assert "str" in names
        assert "name" in names

    def test_collect_names_class_def(self):
        """Test collecting names from a class definition."""
        code = """
class MyClass:
    def method(self, x: int) -> None:
        pass
"""
        tree = ast.parse(code)
        names = collect_names.collect_names(tree)
        assert "int" in names or len(names) >= 0

    def test_collect_names_annotated_assignment(self):
        """Test collecting names from annotated assignments."""
        code = "x: int = 5"
        tree = ast.parse(code)
        names = collect_names.collect_names(tree)
        assert "int" in names

    def test_collect_names_forward_reference(self):
        """Test collecting names from forward references."""
        code = 'def func(x: "ForwardRef") -> None: pass'
        tree = ast.parse(code)
        names = collect_names.collect_names(tree)
        assert "ForwardRef" in names or len(names) >= 0

    def test_collect_names_attribute_access(self):
        """Test collecting names from attribute access."""
        code = """
import os
path = os.path.join("a", "b")
"""
        tree = ast.parse(code)
        names = collect_names.collect_names(tree)
        assert "os" in names

    def test_collect_names_multiple_imports(self):
        """Test collecting names from multiple imports."""
        code = """
from typing import Dict, List, Optional
x: Dict[str, List[int]] = {}
"""
        tree = ast.parse(code)
        names = collect_names.collect_names(tree)
        assert "Dict" in names or "typing" in names

    def test_collect_names_builtin_types(self):
        """Test collecting builtin type names."""
        code = """
def func(x: int, y: str, z: bool) -> list:
    pass
"""
        tree = ast.parse(code)
        names = collect_names.collect_names(tree)
        assert "int" in names
        assert "str" in names
        assert "bool" in names
        assert "list" in names

    def test_collect_names_async_function(self):
        """Test collecting names from async functions."""
        code = """
async def async_func(x: int) -> str:
    return str(x)
"""
        tree = ast.parse(code)
        names = collect_names.collect_names(tree)
        assert "int" in names
        assert "str" in names

    def test_collect_names_vararg_kwarg(self):
        """Test collecting names from *args and **kwargs."""
        code = """
def func(*args: int, **kwargs: str) -> None:
    pass
"""
        tree = ast.parse(code)
        names = collect_names.collect_names(tree)
        assert "int" in names
        assert "str" in names

    def test_collect_names_posonly_args(self):
        """Test collecting names from positional-only arguments."""
        code = """
def func(a: int, /, b: str) -> None:
    pass
"""
        tree = ast.parse(code)
        names = collect_names.collect_names(tree)
        assert "int" in names
        assert "str" in names


class TestDeduplicateImports:
    """Test the deduplicate_imports module."""

    def test_deduplicate_imports_no_duplicates(self):
        """Test that code with no duplicates is unchanged."""
        code = """
import os
import sys
"""
        tree = ast.parse(code)
        result = deduplicate_imports.deduplicate_imports(tree)
        assert len(result.body) == 2

    def test_deduplicate_imports_simple_duplicate(self):
        """Test removing a simple duplicate import."""
        code = """
import os
import os
"""
        tree = ast.parse(code)
        result = deduplicate_imports.deduplicate_imports(tree)
        assert len(result.body) == 1

    def test_deduplicate_imports_from_import_duplicate(self):
        """Test removing duplicate from imports."""
        code = """
from typing import Dict
from typing import Dict
"""
        tree = ast.parse(code)
        result = deduplicate_imports.deduplicate_imports(tree)
        assert len(result.body) == 1

    def test_deduplicate_imports_mixed_import_forms(self):
        """Test deduplication with mixed import forms."""
        code = """
import os
from os import path
import os
"""
        tree = ast.parse(code)
        result = deduplicate_imports.deduplicate_imports(tree)
        # Should keep os from the last import
        assert len(result.body) >= 1

    def test_deduplicate_imports_multiple_names_single_statement(self):
        """Test with multiple imports in one statement."""
        code = """
from typing import Dict, List
from typing import List, Optional
"""
        tree = ast.parse(code)
        result = deduplicate_imports.deduplicate_imports(tree)
        assert isinstance(result, ast.Module)

    def test_deduplicate_imports_alias_handling(self):
        """Test deduplication with import aliases."""
        code = """
import numpy as np
import numpy as np
"""
        tree = ast.parse(code)
        result = deduplicate_imports.deduplicate_imports(tree)
        assert len(result.body) == 1

    def test_deduplicate_imports_from_alias(self):
        """Test deduplication with from import aliases."""
        code = """
from os import path as p
from os import path as p
"""
        tree = ast.parse(code)
        result = deduplicate_imports.deduplicate_imports(tree)
        assert len(result.body) == 1

    def test_deduplicate_imports_star_import(self):
        """Test deduplication of star imports."""
        code = """
from typing import *
from typing import *
"""
        tree = ast.parse(code)
        result = deduplicate_imports.deduplicate_imports(tree)
        assert len(result.body) == 1

    def test_deduplicate_imports_different_modules(self):
        """Test that imports from different modules aren't deduplicated."""
        code = """
from os import path
from pathlib import Path
"""
        tree = ast.parse(code)
        result = deduplicate_imports.deduplicate_imports(tree)
        assert len(result.body) == 2

    def test_deduplicate_imports_keeps_last(self):
        """Test that the last import is kept."""
        code = """
import os
import os
"""
        tree = ast.parse(code)
        result = deduplicate_imports.deduplicate_imports(tree)
        assert len(result.body) == 1

    def test_deduplicate_imports_complex_case(self):
        """Test complex deduplication case."""
        code = """
import os
from typing import Dict
import sys
from typing import Dict, List
import os
"""
        tree = ast.parse(code)
        result = deduplicate_imports.deduplicate_imports(tree)
        assert isinstance(result, ast.Module)
        assert len(result.body) >= 2


class TestNormalizeNames:
    """Test the normalize_names module."""

    def test_normalize_cython_bool(self):
        """Test normalizing Cython bint to bool."""
        code = "def func(x: bint) -> bint: pass"
        tree = ast.parse(code)
        result = normalize_names.normalize_names(tree)
        result_str = ast.unparse(result)
        assert "bool" in result_str

    def test_normalize_cython_unicode(self):
        """Test normalizing Cython unicode to str."""
        code = "def func(x: unicode) -> unicode: pass"
        tree = ast.parse(code)
        result = normalize_names.normalize_names(tree)
        result_str = ast.unparse(result)
        assert "str" in result_str

    def test_normalize_preserves_other_names(self):
        """Test that non-Cython names are preserved."""
        code = "def func(x: MyClass) -> int: pass"
        tree = ast.parse(code)
        result = normalize_names.normalize_names(tree)
        result_str = ast.unparse(result)
        assert "MyClass" in result_str
        assert "int" in result_str


class TestSortImports:
    """Test the sort_imports module."""

    def test_sort_imports_already_sorted(self):
        """Test sorting already sorted imports."""
        code = """
import os
import sys
from typing import Dict
"""
        result = sort_imports.sort_imports(code)
        assert result == code

    def test_sort_imports_unsorted(self):
        """Test sorting unsorted imports."""
        code = """
from typing import Dict
import sys
import os
"""
        expected = """
import os
import sys
from typing import Dict
"""
        result = sort_imports.sort_imports(code)
        assert result == expected

    def test_sort_imports_preserves_code(self):
        """Test that sorting preserves non-import code."""
        code = """
import os
def hello():
    pass
import sys
"""
        result = sort_imports.sort_imports(code)
        assert "hello" in result


class TestTrimImports:
    """Test the trim_imports module."""

    def test_trim_imports_removes_unused(self):
        """Test that unused imports are removed."""
        code = """
import os
import sys

def hello():
    print("hello")
"""
        tree = ast.parse(code)
        result = trim_imports.trim_imports(tree, {"hello", "print"})
        result_str = ast.unparse(result)

        assert "os" not in result_str
        assert "sys" not in result_str

    def test_trim_imports_keeps_used(self):
        """Test that used imports are kept."""
        code = """
import os

def get_cwd():
    return os.getcwd()
"""
        tree = ast.parse(code)
        result = trim_imports.trim_imports(tree, {"get_cwd", "os"})
        result_str = ast.unparse(result)
        assert "os" in result_str

    def test_trim_imports_from_import(self):
        """Test trimming from imports."""
        code = """
from typing import Dict, List

x: Dict[str, int] = {}
"""
        tree = ast.parse(code)
        result = trim_imports.trim_imports(tree, {"x", "Dict", "str", "int"})
        result_str = ast.unparse(result)
        assert "Dict" in result_str
        assert "List" not in result_str



"""Tests for conversion module utilities."""

from __future__ import annotations


from stubgen_pyx.conversion import conversion_utils


class TestLinesCache:
    """Test the _LinesCache class."""

    def test_lines_cache_initialization(self):
        """Test creating a cache."""
        cache = conversion_utils._LinesCache()
        assert cache.source is None
        assert cache.lines == []

    def test_lines_cache_set_source(self):
        """Test setting source in the cache."""
        cache = conversion_utils._LinesCache()
        code = "line1\nline2\nline3"
        cache.source = code
        assert cache.source == code
        assert len(cache.lines) >= 3

    def test_lines_cache_reuse_same_source(self):
        """Test that same source doesn't trigger resplit."""
        cache = conversion_utils._LinesCache()
        code = "line1\nline2"
        cache.source = code
        old_lines = cache.lines
        cache.source = code
        assert cache.lines == old_lines or len(cache.lines) > 0

    def test_lines_cache_different_source(self):
        """Test that different source causes resplit."""
        cache = conversion_utils._LinesCache()
        cache.source = "line1\nline2"
        first_count = len(cache.lines)
        cache.source = "line1\nline2\nline3\nline4"
        assert len(cache.lines) > first_count


class TestDocstringToString:
    """Test the docstring_to_string function."""

    def test_docstring_single_line(self):
        """Test converting a single-line docstring."""
        result = conversion_utils.docstring_to_string("Simple docstring")
        assert '"""' in result
        assert "Simple docstring" in result

    def test_docstring_multiline(self):
        """Test converting a multi-line docstring."""
        docstring = "First line\nSecond line\nThird line"
        result = conversion_utils.docstring_to_string(docstring)
        assert '"""' in result

    def test_docstring_with_triple_quotes(self):
        """Test converting docstring that contains triple quotes."""
        docstring = 'Contains """ inside'
        result = conversion_utils.docstring_to_string(docstring)
        assert '"""' in result

    def test_docstring_with_indentation(self):
        """Test converting docstring with indentation."""
        docstring = "Line 1\n    Indented line\nLine 3"
        result = conversion_utils.docstring_to_string(docstring)
        assert '"""' in result

    def test_docstring_empty(self):
        """Test converting an empty docstring."""
        result = conversion_utils.docstring_to_string("")
        assert '"""' in result


class TestUnparseExpr:
    """Test the unparse_expr function."""

    def test_unparse_none_node(self):
        """Test unparsing a None expression."""
        result = conversion_utils.unparse_expr(None)
        assert result is None

    def test_unparse_simple_value(self):
        """Test that complex expressions become '...'."""
        # We can't directly create Cython nodes easily, so we test the behavior
        result = conversion_utils.unparse_expr(None)
        assert result is None


class TestGetEnumNames:
    """Test the get_enum_names function."""

    def test_get_enum_names_requires_cython_node(self):
        """Test that get_enum_names requires proper Cython node."""
        # This test verifies the function exists and is callable
        # Actual testing requires Cython AST nodes which are complex to create
        assert callable(conversion_utils.get_enum_names)


class TestGetDecorators:
    """Test the get_decorators function."""

    def test_get_decorators_is_callable(self):
        """Test that get_decorators is callable."""
        assert callable(conversion_utils.get_decorators)


class TestGetBases:
    """Test the get_bases function."""

    def test_get_bases_is_callable(self):
        """Test that get_bases is callable."""
        assert callable(conversion_utils.get_bases)


class TestGetMetaclass:
    """Test the get_metaclass function."""

    def test_get_metaclass_is_callable(self):
        """Test that get_metaclass is callable."""
        assert callable(conversion_utils.get_metaclass)


class TestGetSource:
    """Test the get_source function."""

    def test_get_source_is_callable(self):
        """Test that get_source is callable."""
        assert callable(conversion_utils.get_source)



"""StubgenPyx converts .pyx files to .pyi files."""

from __future__ import annotations

from dataclasses import dataclass, field
import glob
import logging
from typing import Iterable
from pathlib import Path

from .config import StubgenPyxConfig
from .analysis.visitor import ModuleVisitor
from .conversion.converter import Converter
from .builders.builder import Builder
from .parsing.parser import parse_pyx, path_to_module_name
from .postprocessing.pipeline import postprocessing_pipeline
from .models.pyi_elements import PyiImport, PyiModule


logger = logging.getLogger(__name__)


@dataclass
class ConversionResult:
    """Result of converting a single .pyx file to .pyi.

    Attributes:
        success: Whether the conversion completed without error.
        pyx_file: Path to the source .pyx file.
        pyi_file: Path to the generated .pyi file.
        error: Exception if conversion failed, otherwise None.
    """

    success: bool
    pyx_file: Path
    pyi_file: Path
    error: Exception | None = None

    @property
    def status_message(self) -> str:
        """Human-readable status summary."""
        if self.success:
            return f"Converted {self.pyx_file} to {self.pyi_file}"
        return f"Failed to convert {self.pyx_file}: {self.error}"


@dataclass
class StubgenPyx:
    """Primary API for converting Cython .pyx files to .pyi stub files.

    Parses Cython source, extracts types, and generates type stubs with optional
    postprocessing (import normalization, trimming, etc.).

    Attributes:
        converter: Converts Cython AST visitors to PyiElements.
        builder: Builds .pyi text from PyiElements.
        config: Configuration controlling generation behavior.
    """

    converter: Converter = field(default_factory=Converter)
    builder: Builder = field(default_factory=Builder)
    config: StubgenPyxConfig = field(default_factory=StubgenPyxConfig)

    def convert_str(
        self, pyx_str: str, pxd_str: str | None = None, pyx_path: Path | None = None
    ) -> str:
        """Convert Cython source strings to .pyi stub code.

        Args:
            pyx_str: The source Cython code.
            pxd_str: Optional companion .pxd file content to merge.
            pyx_path: Optional file path for context and error messages.

        Returns:
            Generated .pyi stub file content.

        Raises:
            Various exceptions from parsing, conversion, or building.
        """
        module = self.compile_str_to_module(pyx_str, pxd_str, pyx_path)
        content = self.builder.build_module(module)
        return postprocessing_pipeline(content, self.config, pyx_path).strip()

    def compile_str_to_module(
        self, pyx_str: str, pxd_str: str | None = None, pyx_path: Path | None = None
    ) -> PyiModule:
        """Compile Cython source strings into a PyiModule

        Args:
            pyx_str: The source Cython code.
            pxd_str: Optional companion .pxd file content to merge.
            pyx_path: Optional file path for context and error messages.

        Returns:
            PyiModule instance as the result of the compilation.

        Raises:
            Various exceptions from parsing, conversion, or building.
        """
        self.builder.include_private = self.config.include_private

        module_name = path_to_module_name(pyx_path) if pyx_path else None
        parse_result = parse_pyx(pyx_str, module_name=module_name, pyx_path=pyx_path)

        module_visitor = ModuleVisitor(node=parse_result.source_ast)
        module = self.converter.convert_module(module_visitor, parse_result.source)

        if pxd_str and not self.config.no_pxd_to_stubs:
            pxd_parse_result = parse_pyx(
                pxd_str, module_name=module_name, pyx_path=pyx_path
            )
            pxd_visitor = ModuleVisitor(node=pxd_parse_result.source_ast)
            pxd_module = self.converter.convert_module(
                pxd_visitor, pxd_parse_result.source
            )

            extra_imports = pxd_module.imports
            extra_enums = pxd_module.scope.enums
            extra_classes = pxd_module.scope.classes
            extra_assignments = pxd_module.scope.assignments
        else:
            extra_imports = []
            extra_enums = []
            extra_classes = []
            extra_assignments = []

        module.scope.enums += extra_enums
        module.scope.assignments += extra_assignments
        module.imports += extra_imports

        module_class_names = {
            class_element.name for class_element in module.scope.classes
        }

        module.scope.classes += [
            extra_class
            for extra_class in extra_classes
            if extra_class.name not in module_class_names
        ]

        module.imports.append(
            PyiImport(
                statement="from __future__ import annotations",
            )
        )

        return module

    def resolve_glob(self, pyx_file_pattern: str) -> Iterable[Path]:
        """Resolve given glob pattern.

        Args:
            pyx_file_pattern: Glob pattern (e.g., "**/*.pyx", "src/*.pyx").

        Returns:
            Iterable of Path to the resolved file names
        """

        # Note: Path.glob is not suitable with patterns containing absolute
        # paths, so glob is much better here
        pyx_files = glob.glob(pyx_file_pattern, recursive=True)

        if len(pyx_files) == 0:
            logger.warning(f"No files matched pattern: {pyx_file_pattern}")
        else:
            logger.info(f"Found {len(pyx_files)} file(s) to convert")
        return (Path(p) for p in pyx_files)

    def convert_glob(
        self,
        pyx_file_pattern: str,
        output_dir: Path | None = None,
        dry_run: bool = False,
    ) -> list[ConversionResult]:
        """Convert multiple .pyx files matching a glob pattern.

        Args:
            pyx_file_pattern: Glob pattern (e.g., "**/*.pyx", "src/*.pyx").
            output_dir: Optional output directory for .pyi files. If None,
                .pyi files are placed next to their source files.
            dry_run: If True, no files are actually created.

        Returns:
            List of ConversionResult objects with status for each file.
        """
        pyx_files = self.resolve_glob(pyx_file_pattern)
        return self.convert_multiple_files(
            pyx_files, output_dir=output_dir, dry_run=dry_run
        )

    def convert_multiple_files(
        self,
        pyx_file_paths: Iterable[Path],
        output_dir: Path | None = None,
        dry_run: bool = False,
    ) -> list[ConversionResult]:
        """Convert multiple .pyx files, each possibly merging a companion .pxd file.

        Reads the .pyx files and companions .pxd (if they exist), converts them,
        and writes the resulting .pyi file.

        Args:
            pyx_file_paths: Paths to the input .pyx files.
            output_dir: Optional output directory for .pyi files. If None,
                .pyi files are placed next to their source files.
            dry_run: If True, no files are actually created.

        Returns:
            ConversionResult with success status and any error details.
        """
        results: list[ConversionResult] = []

        for pyx_path in pyx_file_paths:
            pyi_path = (
                output_dir / pyx_path.with_suffix(".pyi").name if output_dir else None
            )
            result = self.convert_single_file(pyx_path, pyi_path, dry_run)
            results.append(result)

            if self.config.verbose or not result.success:
                logger.info(result.status_message)

        return results

    def convert_single_file(
        self,
        pyx_file_path: Path,
        pyi_file_path: Path | None = None,
        dry_run: bool = False,
    ) -> ConversionResult:
        """Convert a single .pyx file, optionally merging a companion .pxd file.

        Reads the .pyx file and companion .pxd (if it exists), converts them,
        and writes the resulting .pyi file.

        Args:
            pyx_file_path: Path to the input .pyx file.
            pyi_file_path: Path to write the output .pyi file. If None,
                defaults to the same location as the .pyx file with .pyi extension.
            dry_run: If True, no files are actually created.

        Returns:
            ConversionResult with success status and any error details.
        """
        pyi_file_path = pyi_file_path or pyx_file_path.with_suffix(".pyi")
        try:
            logger.debug(f"Converting '{pyx_file_path}' to '{pyi_file_path}'")

            try:
                pyx_str = pyx_file_path.read_text(encoding="utf-8")
            except UnicodeDecodeError as e:
                raise ValueError(f"File encoding error in {pyx_file_path}: {e}") from e
            except FileNotFoundError as e:
                raise ValueError(f"File not found: {pyx_file_path}") from e

            pxd_str = None
            pxd_file_path = pyx_file_path.with_suffix(".pxd")
            if (
                pxd_file_path.exists()
                and not self.config.no_pxd_to_stubs
                and pxd_file_path != pyx_file_path
            ):
                logger.debug(f"Found pxd file: {pxd_file_path}")
                try:
                    pxd_str = pxd_file_path.read_text(encoding="utf-8")
                except UnicodeDecodeError as e:
                    logger.warning(f"Could not read .pxd file {pxd_file_path}: {e}")

            pyi_content = self.convert_str(
                pyx_str=pyx_str,
                pxd_str=pxd_str,
                pyx_path=pyx_file_path,
            )

            if not dry_run:
                try:
                    pyi_file_path.write_text(pyi_content, encoding="utf-8")
                    logger.debug(f"Wrote pyi file: {pyi_file_path}")
                except IOError as e:
                    raise IOError(f"Failed to write {pyi_file_path}: {e}") from e
            else:
                logger.info(f"Would create output file: {pyi_file_path}")

            return ConversionResult(
                success=True,
                pyx_file=pyx_file_path,
                pyi_file=pyi_file_path,
            )

        except Exception as e:
            logger.debug(
                f"Error during conversion: {type(e).__name__}: {e}", exc_info=True
            )

            if not self.config.continue_on_error:
                raise

            return ConversionResult(
                success=False,
                pyx_file=pyx_file_path,
                pyi_file=pyi_file_path,
                error=e,
            )



"""Cython module parser using Cython compiler internals.

Parses Cython code after preprocessing for directives and pragmas.
See `preprocess.py` for preprocessing details.
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
import typing

from Cython.Compiler.TreeFragment import parse_from_strings, StringParseContext
from Cython.Compiler import Errors
from Cython.Compiler.ModuleNode import ModuleNode

from .file_parsing import file_parsing_preprocess
from .preprocess import preprocess

Errors.init_thread()


@dataclass
class ParsedSource:
    """Parsed source code and its AST.

    Attributes:
        source: The preprocessed source code text.
        source_ast: The Cython compiler AST.
    """

    source: str
    source_ast: ModuleNode


_DEFAULT_MODULE_NAME = "__pyx_module__"


def parse_pyx(
    source: str, module_name: str | None = None, pyx_path: Path | None = None
) -> ParsedSource:
    """Parse Cython source code.

    Applies file and string preprocessing, then parses with Cython compiler.

    Args:
        source: Cython source code string.
        module_name: Optional module name for error messages.
        pyx_path: Optional file path for context and preprocessing.

    Returns:
        ParsedSource with preprocessed code and AST.
    """
    module_name = module_name or _DEFAULT_MODULE_NAME

    if pyx_path:
        source = file_parsing_preprocess(pyx_path, source)
        module_name = path_to_module_name(pyx_path)

    return _parse_str(source, module_name)


def _parse_str(source: str, module_name: str) -> ParsedSource:
    """Internal: parse preprocessed source with Cython compiler."""
    context = StringParseContext(module_name, cpp=True)

    source = preprocess(source)

    ast = parse_from_strings(module_name, source, context=context)
    ast = typing.cast("ModuleNode", ast)

    parsed = ParsedSource(source, ast)
    return parsed


def _normalize_part(part: str) -> str:
    """Replace special characters with underscores for module names."""
    return part.replace("-", "_").replace(".", "_").replace(" ", "_")


def path_to_module_name(path: Path) -> str:
    """Convert a file path to a Python module name.

    Handles path separators and special characters for debugging context.
    """
    return ".".join([_normalize_part(part) for part in path.with_suffix("").parts])



"""AST visitors for collecting Cython nodes during tree traversal."""

from __future__ import annotations

from dataclasses import dataclass, field

from Cython.Compiler import Nodes, ExprNodes, ModuleNode
from Cython.Compiler.Visitor import TreeVisitor


@dataclass
class ScopeVisitor(TreeVisitor):
    """Traverses and collects Cython AST nodes in a scope.

    Attributes:
        node: The root node to visit.
        assignments: Collected variable assignments and annotated names.
        py_functions: Collected Python (def) function definitions.
        cdef_functions: Collected Cython (cdef) function definitions.
        classes: Collected class definitions.
        enums: Collected enum definitions.
    """

    node: Nodes.Node
    assignments: list[Nodes.SingleAssignmentNode] = field(
        default_factory=list, init=False
    )
    py_functions: list[Nodes.DefNode] = field(default_factory=list, init=False)
    cdef_functions: list[Nodes.CFuncDefNode] = field(default_factory=list, init=False)
    classes: list[ClassVisitor] = field(default_factory=list, init=False)
    enums: list[Nodes.CEnumDefNode] = field(default_factory=list, init=False)

    def __post_init__(self):
        super().__init__()
        self.visitchildren(self.node)

    def visit_Node(self, node):
        """Generic node visitor."""
        return node

    def visit_CEnumDefNode(self, node):
        """Collect public enums."""
        if not node.name:
            return node
        self.enums.append(node)
        return node

    def visit_StatListNode(self, node):
        """Traverse statement lists."""
        self.visitchildren(node)
        return node

    def visit_CDefExternNode(self, node):
        """Propagate visit from "cdef extern" nodes.

        `cdef extern from "<header.h>"` allows for declaration of enumerations
        to be wrapped when `cpdef enum` is used. Here visiting is propagated
        down to the children to search for those enums.
        """
        self.visitchildren(node)
        return node

    def visit_SingleAssignmentNode(self, node):
        """Collect non-import assignments."""
        if isinstance(node.rhs, ExprNodes.ImportNode):
            return node
        if isinstance(node.lhs, ExprNodes.NameNode):
            self.assignments.append(node)
        return node

    def visit_ExprStatNode(self, node):
        """Collect annotated name expressions."""
        if (
            isinstance(node.expr, ExprNodes.NameNode)
            and node.expr.annotation is not None
        ):
            self.assignments.append(node)
        return node

    def visit_PyClassDefNode(self, node):
        """Collect Python class definitions."""
        self.classes.append(ClassVisitor(node=node))
        return node

    def visit_CClassDefNode(self, node):
        """Collect Cython extension type (cdef class) definitions."""
        self.classes.append(ClassVisitor(node=node))
        return node

    def visit_DefNode(self, node):
        """Collect Python function definitions."""
        self.py_functions.append(node)
        return node

    def visit_CFuncDefNode(self, node):
        """Collect Python-visible (cpdef) function definitions."""
        if not node.declarator.overridable:
            return node
        self.cdef_functions.append(node)
        return node

    def visit_CTypeDefNode(self, node):
        """Collect simple Cython type definitions."""
        if isinstance(node.declarator, Nodes.CNameDeclaratorNode):
            self.assignments.append(node)
        return node


@dataclass
class ImportVisitor(TreeVisitor):
    """Visits and collects Cython import nodes in a scope."""

    node: Nodes.Node
    """The node to visit."""

    imports: list[Nodes.Node] = field(default_factory=list, init=False)
    """A list of collected import statements."""

    def __post_init__(self):
        super().__init__()
        self.visitchildren(self.node)

    def visit_Node(self, node):
        """Default visitor for generic nodes."""
        return node

    def visit_StatListNode(self, node):
        """Visits statement list nodes and their children."""
        self.visitchildren(node)
        return node

    def visit_CImportNode(self, node):
        """Visits cimport nodes."""
        self.imports.append(node)
        return node

    def visit_CImportStatNode(self, node):
        """Visits cimport statement nodes."""
        self.imports.append(node)
        return node

    def visit_ImportNode(self, node):
        """Visits import nodes."""
        self.imports.append(node)
        return node

    def visit_FromImportNode(self, node):
        """Visits from...import nodes."""
        self.imports.append(node)
        return node

    def visit_FromImportStatNode(self, node):
        """Visits from...import statement nodes."""
        self.imports.append(node)
        return node

    def visit_FromCImportStatNode(self, node):
        """Visits from...cimport statement nodes."""
        self.imports.append(node)
        return node

    def visit_ImportStatNode(self, node):
        """Visits import statement nodes."""
        self.imports.append(node)
        return node

    def visit_SingleAssignmentNode(self, node):
        if isinstance(node.rhs, ExprNodes.ImportNode):
            self.imports.append(node)
            return node
        return node

    def visit_IfStatNode(self, node):
        """Pass through `if typing.TYPE_CHECKING: ` and `if TYPE_CHECKING: ` blocks"""
        for clause in node.if_clauses:
            condition_name = _collect_attribute(clause.condition)
            if condition_name in ("TYPE_CHECKING", "typing.TYPE_CHECKING"):
                self.visitchildren(clause)


@dataclass
class ModuleVisitor:
    """Visits and collects Cython module nodes in a scope."""

    node: ModuleNode.ModuleNode
    """The node to visit."""

    import_visitor: ImportVisitor = field(init=False)
    """A visitor for collecting import nodes."""

    scope: ScopeVisitor = field(init=False)
    """A visitor for collecting scope nodes."""

    def __post_init__(self):
        self.import_visitor = ImportVisitor(node=self.node)
        self.scope = ScopeVisitor(node=self.node)


@dataclass
class ClassVisitor:
    """Visits and collects Cython class nodes in a scope."""

    node: Nodes.CClassDefNode | Nodes.PyClassDefNode
    """The node to visit."""

    scope: ScopeVisitor = field(init=False)
    """A visitor for collecting scope nodes."""

    def __post_init__(self):
        self.scope = ScopeVisitor(node=self.node)


def _collect_attribute(node) -> str:
    names = []
    attribute = node

    while isinstance(attribute, ExprNodes.AttributeNode):
        names.append(attribute.attribute)
        attribute = attribute.obj

    if isinstance(attribute, ExprNodes.NameNode):
        names.append(attribute.name)  # type: ignore

    names.reverse()

    name = ".".join(names)
    return name



"""Normalizes Cython type names to Python equivalents in .pyi files."""

from __future__ import annotations

import ast
from dataclasses import dataclass


def normalize_names(tree: ast.AST) -> ast.AST:
    """Replace Cython type names with their Python equivalents in an AST."""
    return _NameNormalizer().visit(tree)


_CYTHON_INTS: tuple[str, ...] = (
    "char",
    "short",
    "Py_UNICODE",
    "Py_UCS4",
    "long",
    "longlong",
    "Py_hash_t",
    "Py_ssize_t",
    "size_t",
    "ssize_t",
    "ptrdiff_t",
    "int64_t",
    "int32_t",
    "int16_t",
    "uint64_t",
    "uint32_t",
    "uint16_t",
)

_CYTHON_FLOATS: tuple[str, ...] = (
    "double",
    "longdouble",
)

_CYTHON_COMPLEXES: tuple[str, ...] = (
    "longdoublecomplex",
    "doublecomplex",
    "floatcomplex",
)

_CYTHON_TRANSLATIONS: dict[str, str] = {
    "bint": "bool",
    "unicode": "str",
    "void": "None",
}

for int_type in _CYTHON_INTS:
    _CYTHON_TRANSLATIONS[int_type] = "int"
for float_type in _CYTHON_FLOATS:
    _CYTHON_TRANSLATIONS[float_type] = "float"
for complex_type in _CYTHON_COMPLEXES:
    _CYTHON_TRANSLATIONS[complex_type] = "complex"


@dataclass
class _NameNormalizer(ast.NodeTransformer):
    """Transform Cython type names to Python equivalents."""

    def visit_Name(self, node: ast.Name) -> ast.Name:
        name = _CYTHON_TRANSLATIONS.get(node.id, node.id)
        return ast.Name(id=name, ctx=node.ctx)



"""Tests for builders module."""

from __future__ import annotations


from stubgen_pyx.builders.builder import Builder
from stubgen_pyx.models.pyi_elements import (
    PyiModule,
    PyiClass,
    PyiFunction,
    PyiAssignment,
    PyiImport,
    PyiScope,
    PyiSignature,
    PyiArgument,
    PyiEnum,
)


class TestBuilder:
    """Test the Builder class."""

    def test_builder_creation_default(self):
        """Test creating a builder with default settings."""
        builder = Builder()
        assert builder.include_private is False

    def test_builder_creation_include_private(self):
        """Test creating a builder that includes private members."""
        builder = Builder(include_private=True)
        assert builder.include_private is True

    def test_is_private_private_name(self):
        """Test identifying private names."""
        builder = Builder()
        assert builder._is_private("_private") is True
        assert builder._is_private("__private") is True

    def test_is_private_dunder_name(self):
        """Test that dunder names are not considered private."""
        builder = Builder()
        assert builder._is_private("__init__") is False
        assert builder._is_private("__str__") is False

    def test_is_private_public_name(self):
        """Test that public names are not private."""
        builder = Builder()
        assert builder._is_private("public") is False
        assert builder._is_private("MyClass") is False

    def test_build_argument_name_only(self):
        """Test building an argument with only a name."""
        builder = Builder()
        arg = PyiArgument("x")
        result = builder.build_argument(arg)
        assert result == "x"

    def test_build_argument_with_annotation(self):
        """Test building an argument with an annotation."""
        builder = Builder()
        arg = PyiArgument("x", annotation="int")
        result = builder.build_argument(arg)
        assert result == "x: int"

    def test_build_argument_with_default(self):
        """Test building an argument with a default value."""
        builder = Builder()
        arg = PyiArgument("x", annotation="int", default="5")
        result = builder.build_argument(arg)
        assert result == "x: int = 5"

    def test_build_argument_with_default_no_annotation(self):
        """Test building an argument with default but no annotation."""
        builder = Builder()
        arg = PyiArgument("x", default="None")
        result = builder.build_argument(arg)
        assert result == "x = None"

    def test_build_signature_simple(self):
        """Test building a simple signature."""
        builder = Builder()
        sig = PyiSignature(args=[PyiArgument("x", annotation="int")])
        result = builder.build_signature(sig)
        assert "(x: int)" in result

    def test_build_signature_multiple_args(self):
        """Test building a signature with multiple arguments."""
        builder = Builder()
        sig = PyiSignature(
            args=[
                PyiArgument("x", annotation="int"),
                PyiArgument("y", annotation="str"),
            ]
        )
        result = builder.build_signature(sig)
        assert "x: int" in result
        assert "y: str" in result

    def test_build_signature_with_return_type(self):
        """Test building a signature with return type."""
        builder = Builder()
        sig = PyiSignature(args=[PyiArgument("x", "int")], return_type="str")
        result = builder.build_signature(sig)
        assert "-> str" in result

    def test_build_signature_with_varargs(self):
        """Test building a signature with *args."""
        builder = Builder()
        sig = PyiSignature(args=[], var_arg=PyiArgument("args", "int"))
        result = builder.build_signature(sig)
        assert "*" in result

    def test_build_signature_with_kwargs(self):
        """Test building a signature with **kwargs."""
        builder = Builder()
        sig = PyiSignature(args=[], kw_arg=PyiArgument("kwargs", "str"))
        result = builder.build_signature(sig)
        assert "**" in result

    def test_build_signature_posonly_args(self):
        """Test building a signature with positional-only arguments."""
        builder = Builder()
        sig = PyiSignature(
            args=[
                PyiArgument("x", "int"),
                PyiArgument("y", "str"),
            ],
            num_posonly_args=1,
        )
        result = builder.build_signature(sig)
        assert "/" in result

    def test_build_signature_kwonly_args(self):
        """Test building a signature with keyword-only arguments."""
        builder = Builder()
        sig = PyiSignature(
            args=[
                PyiArgument("x", "int"),
                PyiArgument("y", "str"),
            ],
            num_kwonly_args=1,
        )
        result = builder.build_signature(sig)
        assert "*" in result

    def test_build_function_simple(self):
        """Test building a simple function."""
        builder = Builder()
        func = PyiFunction(
            name="greet",
            is_async=False,
            signature=PyiSignature(
                args=[PyiArgument("name", annotation="str")], return_type="str"
            ),
        )
        result = builder.build_function(func)
        assert result
        assert "def greet" in result
        assert "(name: str)" in result
        assert "-> str" in result

    def test_build_function_with_docstring(self):
        """Test building a function with a docstring."""
        builder = Builder()
        func = PyiFunction(
            name="greet",
            is_async=False,
            signature=PyiSignature(),
            doc='"""Greet someone."""',
        )
        result = builder.build_function(func)
        assert result
        assert "def greet" in result
        assert "Greet someone" in result

    def test_build_function_async(self):
        """Test building an async function."""
        builder = Builder()
        func = PyiFunction(name="async_greet", signature=PyiSignature(), is_async=True)
        result = builder.build_function(func)
        assert result
        assert "async def" in result

    def test_build_function_with_decorator(self):
        """Test building a function with decorators."""
        builder = Builder()
        func = PyiFunction(
            name="decorated_func",
            is_async=False,
            signature=PyiSignature(),
            decorators=["@property", "@staticmethod"],
        )
        result = builder.build_function(func)
        assert result
        assert "@property" in result
        assert "@staticmethod" in result

    def test_build_function_private_included(self):
        """Test building a private function when include_private is True."""
        builder = Builder(include_private=True)
        func = PyiFunction(
            name="_private_func", is_async=False, signature=PyiSignature()
        )
        result = builder.build_function(func)
        assert result
        assert "_private_func" in result

    def test_build_assignment_simple(self):
        """Test building a simple assignment."""
        builder = Builder()
        assignment = PyiAssignment("x: int = 5")
        result = builder.build_assignment(assignment)
        assert result == "x: int = 5"

    def test_build_import_simple(self):
        """Test building a simple import statement."""
        builder = Builder()
        import_stmt = PyiImport("import os")
        result = builder.build_import(import_stmt)
        assert result == "import os"

    def test_build_class_simple(self):
        """Test building a simple class."""
        builder = Builder()
        cls = PyiClass(name="MyClass", scope=PyiScope())
        result = builder.build_class(cls)
        assert result
        assert "class MyClass" in result
        assert "..." in result

    def test_build_class_with_base(self):
        """Test building a class with a base class."""
        builder = Builder()
        cls = PyiClass(name="MyClass", bases=["BaseClass"], scope=PyiScope())
        result = builder.build_class(cls)
        assert result
        assert "class MyClass(BaseClass" in result

    def test_build_class_with_multiple_bases(self):
        """Test building a class with multiple bases."""
        builder = Builder()
        cls = PyiClass(
            name="MyClass", bases=["BaseClass1", "BaseClass2"], scope=PyiScope()
        )
        result = builder.build_class(cls)
        assert result
        assert "BaseClass1" in result
        assert "BaseClass2" in result

    def test_build_class_with_metaclass(self):
        """Test building a class with a metaclass."""
        builder = Builder()
        cls = PyiClass(name="MyClass", metaclass="MetaClass", scope=PyiScope())
        result = builder.build_class(cls)
        assert result
        assert "metaclass=MetaClass" in result

    def test_build_class_with_docstring(self):
        """Test building a class with a docstring."""
        builder = Builder()
        cls = PyiClass(name="MyClass", doc='"""My class."""', scope=PyiScope())
        result = builder.build_class(cls)
        assert result
        assert "MyClass" in result
        assert "My class" in result

    def test_build_class_with_decorator(self):
        """Test building a class with decorators."""
        builder = Builder()
        cls = PyiClass(name="MyClass", decorators=["@dataclass"], scope=PyiScope())
        result = builder.build_class(cls)
        assert result
        assert "@dataclass" in result

    def test_build_scope_with_assignments(self):
        """Test building a scope with assignments."""
        builder = Builder()
        scope = PyiScope(assignments=[PyiAssignment("x: int")])
        result = builder.build_scope(scope)
        assert result
        assert "x: int" in result

    def test_build_scope_with_functions(self):
        """Test building a scope with functions."""
        builder = Builder()
        scope = PyiScope(
            functions=[PyiFunction("hello", is_async=False, signature=PyiSignature())]
        )
        result = builder.build_scope(scope)
        assert result
        assert "hello" in result

    def test_build_scope_with_private_function_excluded(self):
        """Test that private functions are excluded by default."""
        builder = Builder(include_private=False)
        scope = PyiScope(
            functions=[
                PyiFunction("_private", is_async=False, signature=PyiSignature())
            ]
        )
        result = builder.build_scope(scope)
        assert result
        assert "_private" not in result

    def test_build_scope_with_private_function_included(self):
        """Test that private functions are included when specified."""
        builder = Builder(include_private=True)
        scope = PyiScope(
            functions=[
                PyiFunction("_private", is_async=False, signature=PyiSignature())
            ]
        )
        result = builder.build_scope(scope)
        assert result
        assert "_private" in result

    def test_build_enum_with_name(self):
        """Test building a named enum."""
        builder = Builder()
        enum = PyiEnum(enum_name="Colors", names=["RED", "GREEN", "BLUE"])
        result = builder.build_enum(enum)
        assert result
        assert "class Colors" in result
        assert "RED: int" in result

    def test_build_enum_without_name(self):
        """Test building an unnamed enum."""
        builder = Builder()
        enum = PyiEnum(enum_name=None, names=["RED", "GREEN", "BLUE"])
        result = builder.build_enum(enum)
        assert result
        assert "RED: int" in result
        assert "GREEN: int" in result

    def test_build_enum_empty(self):
        """Test building an empty enum."""
        builder = Builder()
        enum = PyiEnum(enum_name="Empty", names=[])
        result = builder.build_enum(enum)
        assert result is None

    def test_build_module_simple(self):
        """Test building a simple module."""
        builder = Builder()
        module = PyiModule(scope=PyiScope())
        result = builder.build_module(module)
        assert isinstance(result, str)

    def test_build_module_with_docstring(self):
        """Test building a module with a docstring."""
        builder = Builder()
        module = PyiModule(doc='"""Module docstring."""', scope=PyiScope())
        result = builder.build_module(module)
        assert "Module docstring" in result

    def test_build_module_with_imports(self):
        """Test building a module with imports."""
        builder = Builder()
        module = PyiModule(imports=[PyiImport("import os")], scope=PyiScope())
        result = builder.build_module(module)
        assert "import os" in result

    def test_build_module_complete(self):
        """Test building a complete module."""
        builder = Builder()
        module = PyiModule(
            doc='"""A test module."""',
            imports=[PyiImport("from typing import Dict")],
            scope=PyiScope(
                functions=[
                    PyiFunction("test", is_async=False, signature=PyiSignature())
                ]
            ),
        )
        result = builder.build_module(module)
        assert "test module" in result
        assert "import" in result or "Dict" in result



"""Tests for error handling and edge cases in core modules."""

from __future__ import annotations

import tempfile
from pathlib import Path
from unittest.mock import patch
import pytest

from stubgen_pyx.stubgen import StubgenPyx, ConversionResult
from stubgen_pyx.config import StubgenPyxConfig


class TestStubgenErrorHandling:
    """Test error handling in StubgenPyx."""

    def test_convert_file_with_encoding_error(self):
        """Test converting a file with encoding error."""
        with tempfile.TemporaryDirectory() as tmpdir:
            pyx_file = Path(tmpdir) / "bad_encoding.pyx"
            # Write file with invalid UTF-8
            pyx_file.write_bytes(b"\x80\x81\x82")

            config = StubgenPyxConfig(continue_on_error=True)
            stubgen = StubgenPyx(config=config)

            result = stubgen.convert_glob(str(pyx_file))
            assert result[0].success is False

    def test_convert_file_with_file_not_found(self):
        """Test converting a non-existent file."""
        config = StubgenPyxConfig(continue_on_error=True)
        stubgen = StubgenPyx(config=config)

        result = stubgen.convert_glob("__does_not_exist.pyx")
        assert not result  # empty list

    def test_convert_file_with_write_error(self):
        """Test when writing output file fails."""
        with tempfile.TemporaryDirectory() as tmpdir:
            pyx_file = Path(tmpdir) / "test.pyx"
            pyx_file.write_text("def hello(): pass")

            # Create a mock output dir that can't be written to
            with patch.object(
                Path, "write_text", side_effect=IOError("Permission denied")
            ):
                config = StubgenPyxConfig(continue_on_error=True)
                stubgen = StubgenPyx(config=config)

                result = stubgen.convert_glob(str(pyx_file))
                assert result[0].success is False

    def test_convert_glob_with_error_in_middle(self):
        """Test glob conversion when one file fails."""
        with tempfile.TemporaryDirectory() as tmpdir:
            tmppath = Path(tmpdir)

            # Create valid file
            valid_file = tmppath / "valid.pyx"
            valid_file.write_text("def hello(): pass")

            # Create file with bad encoding
            bad_file = tmppath / "bad.pyx"
            bad_file.write_bytes(b"\x80\x81\x82")

            # Create another valid file
            valid_file2 = tmppath / "valid2.pyx"
            valid_file2.write_text("def goodbye(): pass")

            config = StubgenPyxConfig(continue_on_error=True)
            stubgen = StubgenPyx(config=config)

            results = stubgen.convert_glob(str(tmppath / "*.pyx"))

            # Should have results for all files
            assert len(results) >= 2
            successful = [r for r in results if r.success]
            failed = [r for r in results if not r.success]
            assert len(successful) >= 2
            assert len(failed) >= 1

    def test_convert_file_without_error_continuation(self):
        """Test that conversion stops on error when continue_on_error is False."""
        with tempfile.TemporaryDirectory() as tmpdir:
            pyx_file = Path(tmpdir) / "bad_encoding.pyx"
            pyx_file.write_bytes(b"\x80\x81\x82")

            config = StubgenPyxConfig(continue_on_error=False)
            stubgen = StubgenPyx(config=config)

            with pytest.raises(ValueError):
                stubgen.convert_glob(str(pyx_file))

    def test_conversion_result_creation(self):
        """Test ConversionResult dataclass."""
        pyx_file = Path("test.pyx")
        pyi_file = Path("test.pyi")
        result = ConversionResult(pyx_file=pyx_file, pyi_file=pyi_file, success=True)
        assert result.pyx_file == pyx_file
        assert result.pyi_file == pyi_file
        assert result.success is True

    def test_convert_file_with_pxd_file_encoding_error(self):
        """Test when .pxd file has encoding error."""
        with tempfile.TemporaryDirectory() as tmpdir:
            tmppath = Path(tmpdir)

            # Create valid .pyx file
            pyx_file = tmppath / "test.pyx"
            pyx_file.write_text("def hello(): pass")

            # Create .pxd file with bad encoding
            pxd_file = tmppath / "test.pxd"
            pxd_file.write_bytes(b"\x80\x81\x82")

            config = StubgenPyxConfig(continue_on_error=True, no_pxd_to_stubs=False)
            stubgen = StubgenPyx(config=config)

            result = stubgen.convert_glob(str(tmppath / "*.pyx"))
            assert isinstance(result[0], ConversionResult)

    def test_convert_glob_with_no_files_no_error(self):
        """Test glob with no matches doesn't error."""
        with tempfile.TemporaryDirectory() as tmpdir:
            config = StubgenPyxConfig()
            stubgen = StubgenPyx(config=config)

            results = stubgen.convert_glob(str(Path(tmpdir) / "*.pyx"))
            assert results == []

    def test_stubgen_initialization_with_config(self):
        """Test StubgenPyx initialization with custom config."""
        config = StubgenPyxConfig(
            verbose=True, no_sort_imports=True, include_private=True
        )
        stubgen = StubgenPyx(config=config)
        assert stubgen.config.verbose is True
        assert stubgen.config.no_sort_imports is True
        assert stubgen.config.include_private is True

    def test_stubgen_default_config(self):
        """Test StubgenPyx with default config."""
        stubgen = StubgenPyx()
        assert stubgen.config is not None

    def test_convert_str_empty_code(self):
        """Test converting empty code."""
        stubgen = StubgenPyx()
        result = stubgen.convert_str("")
        assert isinstance(result, str)

    def test_convert_str_only_comments(self):
        """Test converting code with only comments."""
        with tempfile.TemporaryDirectory() as tmpdir:
            pyx_file = Path(tmpdir) / "test.pyx"
            code = "# This is just a comment\n# Another comment"
            result = StubgenPyx().convert_str(code, pyx_path=pyx_file)
            assert isinstance(result, str)

    def test_convert_file_with_output_dir(self):
        """Test converting file with custom output directory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            tmppath = Path(tmpdir)
            source_dir = tmppath / "source"
            source_dir.mkdir()
            output_dir = tmppath / "output"
            output_dir.mkdir()

            pyx_file = source_dir / "test.pyx"
            pyx_file.write_text("def hello(): pass")

            stubgen = StubgenPyx()
            (result,) = stubgen.convert_glob(str(pyx_file), output_dir)

            assert result.success is True
            assert (output_dir / "test.pyi").exists()


class TestConversionResultProperties:
    """Test ConversionResult properties."""

    def test_result_success_true(self):
        """Test result with success=True."""
        result = ConversionResult(
            pyx_file=Path("test.pyx"), success=True, pyi_file=Path("test.pyi")
        )
        assert result.success is True
        assert not hasattr(result, "error") or result.error is None

    def test_result_success_false_with_error(self):
        """Test result with success=False and error message."""
        result = ConversionResult(
            pyx_file=Path("test.pyx"),
            pyi_file=Path("test.pyi"),
            success=False,
            error=SyntaxError("Syntax error"),
        )
        assert result.success is False
        assert isinstance(result.error, SyntaxError)

    def test_result_with_pyi_file(self):
        """Test result with both pyx and pyi files."""
        pyx_file = Path("test.pyx")
        pyi_file = Path("test.pyi")
        result = ConversionResult(pyx_file=pyx_file, pyi_file=pyi_file, success=True)
        assert result.pyx_file == pyx_file
        assert result.pyi_file == pyi_file
